use super::actor::{self, MqttActor};
use actix::Addr;
use rumqttc::{Packet, QoS};
use std::{collections::HashSet, time::Duration};
use tokio::sync::oneshot;
use tracing::{debug, trace, warn};

pub(super) struct EventLoopTask {
    event_loop: rumqttc::EventLoop,
    mqtt_addr: Addr<MqttActor>,
    quit_tx: oneshot::Sender<()>,
    unreleased_pubs: HashSet<u16>,
}

impl EventLoopTask {
    pub(super) fn new(
        event_loop: rumqttc::EventLoop,
        mqtt_addr: Addr<MqttActor>,
        quit_tx: oneshot::Sender<()>,
    ) -> Self {
        Self {
            event_loop,
            mqtt_addr,
            quit_tx,
            unreleased_pubs: HashSet::default(),
        }
    }

    fn handle_incoming(&mut self, incoming: Packet) {
        trace!(incoming = debug(&incoming), "MQTT incoming message");
        #[allow(clippy::single_match)]
        match incoming {
            Packet::ConnAck(_) => {
                self.mqtt_addr.do_send(actor::Connected);
            }
            Packet::Publish(publish) => {
                // Only deliver QoS 2 packets once
                let deliver = if publish.qos == QoS::ExactlyOnce {
                    if self.unreleased_pubs.contains(&publish.pkid) {
                        false
                    } else {
                        self.unreleased_pubs.insert(publish.pkid);
                        true
                    }
                } else {
                    true
                };
                if deliver {
                    self.mqtt_addr.do_send(actor::PubRecieve(publish));
                }
            }
            Packet::PubRel(pubrel) => {
                self.unreleased_pubs.remove(&pubrel.pkid);
            }
            _ => {}
        }
    }

    pub(super) async fn run(mut self) {
        use rumqttc::{ConnectionError, Event};
        let reconnect_timeout = Duration::from_secs(5);
        self.event_loop.set_reconnection_delay(reconnect_timeout);
        loop {
            match self.event_loop.poll().await {
                Ok(Event::Incoming(incoming)) => {
                    self.handle_incoming(incoming);
                }
                Ok(Event::Outgoing(outgoing)) => {
                    trace!(outgoing = debug(&outgoing), "MQTT outgoing message");
                }
                Err(ConnectionError::Cancel) => {
                    debug!("MQTT disconnecting");
                    break;
                }
                Err(err) => {
                    warn!("MQTT error, reconnecting: {}", err);
                }
            }
        }
        let _ = self.quit_tx.send(());
    }
}
